/**      Cartman2ddp1.C by Alexander Myasnikov                     **/
/**      Fixed Cartman Tiny block cipher 2                         **/
/**      Cartman encrypts 128 bit block with 256..512 bit key      **/
/**      Gives much better transforming with new key schedule      **/
/**      Freeware for any use,  non-patented, opensource cipher    **/
/**      WEB:       www.darksoftware.narod.ru                      **/


#include <mem.h>
#include <stdlib.h>

#define u32             unsigned long
#define u8              unsigned char
#define keysizebits     512
#define keysizebytes    keysizebits / 8
#define keysize         keysizebytes / 4
#define subkeysize      keysize / 2
#define ror32(N, R)    _lrotr(N, R)
#define rol32(N, R)    _lrotl(N, R)


u32 idx_box1[subkeysize];
u32 idx_box2[subkeysize];
u32 rot_box2[subkeysize];
u32 rot_box3[subkeysize];
u32 rot_box[keysize];
u32 key[keysize];
u32 subkey1[subkeysize];
u32 subkey2[subkeysize];
u32 idx1null;
u32 idx2null;
u32 idx3null;


#include "SBox.h"

void
makebox ()
{
  u32 i;

  idx1null = 0;
  idx2null = 0;
  idx3null = 0;


  for (i = 0; i != subkeysize; i++)
    {
      idx_box1[i] =
        (ror32 (key[i] + ((key[i] << 4) ^ (key[i] >> 5)), (i % 31) + 1)) %
        (subkeysize - 1);

      idx1null = (idx1null * (idx1null + 1) >> 1) + key[i];
    }

  for (i = subkeysize; i != keysize; i++)
    {
      idx_box2[(keysize - 1) - i] =
        (rol32 (key[i] + ((key[i] << 4) ^ (key[i] >> 5)), (i % 31) + 1)) %
        (subkeysize - 1);
      idx2null = (idx2null * (idx2null + 1) >> 1) + key[i];
    }

  for (i = 0; i != keysize; i++)
    {
      rot_box[i] =
        ((rol32 (key[i] ^ (i + 1), (((keysize - 1) - i) % 31) + 1)) % 31) + 1;
      idx3null += key[i] * (i + 1);
    }

  idx1null = (idx1null ^ (idx1null >> 16)) % subkeysize;
  idx2null = (idx2null ^ (idx2null >> 16)) % subkeysize;
  idx3null = (idx3null ^ (idx3null >> 16)) % subkeysize;

  for (i = 0; i < subkeysize; i++)
    {
      rot_box2[i] = subkey2[(subkeysize - 1) - i] % 32;
      rot_box3[i] = subkey1[(subkeysize - 1) - i] % 32;
    }

}

void __stdcall __export
setup (u32 * k)
{
  memmove (&key[0], &k[0], keysizebytes);

  memmove (&subkey1[0], &key[0], subkeysize * 4);
  memmove (&subkey2[0], &key[subkeysize], subkeysize * 4);

  makebox ();

}

void __stdcall __export
killkey ()
{
  memset (key, 0, keysizebytes);
  memset (key, 0xff, keysizebytes);
}

void __stdcall __export
crypt (u32 * v)
{
  u32 i, k1, k2, k3, k4, rot1, rot2, idx1, idx2, n, v0, v1, j, m, s, t;


  for (n = 0; n != subkeysize; n++)
    {
      idx1 = idx_box1[n];
      idx2 = idx_box2[n];

      for (i = 0; i != subkeysize; i++)
        {

          s = v[0] ^ v[1] ^ v[2] ^ v[3];


          for (j = 0; j < 4; j++)
            {
              m = (s >> (8 * j)) & 3;
              t = v[j];
              v[j] = v[m];
              v[m] = t;
            }

          v[0] =
            (SBox[idx1][idx2][(u8) (v[0] >> 0)] << 0) +
            (SBox[idx1][idx2][(u8) (v[0] >> 8)] << 8) +
            (SBox[idx1][idx2][(u8) (v[0] >> 16)] << 16) +
            (SBox[idx1][idx2][(u8) (v[0] >> 24)] << 24);

          v[1] =
            (SBox[idx1 + 1][idx2 + 1][(u8) (v[1] >> 0)] << 0) +
            (SBox[idx1 + 1][idx2 + 1][(u8) (v[1] >> 8)] << 8) +
            (SBox[idx1 + 1][idx2 + 1][(u8) (v[1] >> 16)] << 16) +
            (SBox[idx1 + 1][idx2 + 1][(u8) (v[1] >> 24)] << 24);

          v[2] =
            (SBox[idx1][idx2 + 1][(u8) (v[2] >> 0)] << 0) +
            (SBox[idx1][idx2 + 1][(u8) (v[2] >> 8)] << 8) +
            (SBox[idx1][idx2 + 1][(u8) (v[2] >> 16)] << 16) +
            (SBox[idx1][idx2 + 1][(u8) (v[2] >> 24)] << 24);

          v[3] =
            (SBox[idx1 + 1][idx2][(u8) (v[3] >> 0)] << 0) +
            (SBox[idx1 + 1][idx2][(u8) (v[3] >> 8)] << 8) +
            (SBox[idx1 + 1][idx2][(u8) (v[3] >> 16)] << 16) +
            (SBox[idx1 + 1][idx2][(u8) (v[3] >> 24)] << 24);

          k1 = subkey1[idx1];
          k2 = subkey1[idx1 + 1];
          k3 = subkey2[idx2];
          k4 = subkey2[idx2 + 1];

          rot1 = rot_box[idx1 + idx2];
          rot2 = rot_box[idx1 + idx2 + 1];

          if (idx1 == idx1null)
            {
              v[0] +=
                (v[3] + ror32 ((v[1] + ((v[1] << 6) ^ (v[1] >> 9))), rot1) +
                 k1);
            }
          else
            {
              v[0] +=
                (v[3] +
                 ror32 ((v[1] + ((v[1] << 6) ^ (v[1] >> 9))) + k1, rot1));
            }

          if (idx2 == idx2null)
            {
              v[1] +=
                (v[2] + ror32 ((v[0] + ((v[0] << 6) ^ (v[0] >> 9))), rot2) +
                 k2);
            }
          else
            {
              v[1] +=
                (v[2] +
                 ror32 ((v[0] + ((v[0] << 6) ^ (v[0] >> 9))) + k2, rot2));
            }


          if (i == idx3null)
            {
              v[2] +=
                rol32 (((v[0] + ((v[0] << 4) ^ (v[0] >> 5))) + k3) * 9,
                       rot_box2[idx1]);
              v[3] +=
                rol32 (((v[1] + ((v[1] << 4) ^ (v[1] >> 5))) + k4) * 9,
                       rot_box3[idx2]);
            }
          else
            {
              v[2] ^=
                rol32 (((v[0] + ((v[0] << 4) ^ (v[0] >> 5))) + k3) * 9,
                       rot_box2[idx1]);
              v[3] ^=
                rol32 (((v[1] + ((v[1] << 4) ^ (v[1] >> 5))) + k4) * 9,
                       rot_box3[idx2]);
            }


          v0 = v[0];
          v1 = v[1];
          v[0] = v[2];
          v[1] = v[3];
          v[2] = v0;
          v[3] = v1;

          if (idx1 != (subkeysize - 2))
            {
              idx1++;
            }
          else
            {
              idx1 = 0;
            }

          if (idx2 != (subkeysize - 2))
            {
              idx2++;
            }
          else
            {
              idx2 = 0;
            }
        }
    }
}

void __stdcall __export
decrypt (u32 * v)
{
  u32 k1, k2, k3, k4, rot1, rot2, idx1, idx2, v0, v1, m, s, t;

  long n, i, j;

  for (n = subkeysize - 1; n != -1; n--)

    {
      idx1 = idx_box1[n];
      idx2 = idx_box2[n];

      for (i = subkeysize - 1; i != -1; i--)
        {
          k1 = subkey1[idx1];
          k2 = subkey1[idx1 + 1];
          k3 = subkey2[idx2];
          k4 = subkey2[idx2 + 1];

          rot1 = rot_box[idx1 + idx2];
          rot2 = rot_box[idx1 + idx2 + 1];

          v0 = v[0];
          v1 = v[1];
          v[0] = v[2];
          v[1] = v[3];
          v[2] = v0;
          v[3] = v1;


          if (i == idx3null)
            {
              v[3] -=
                rol32 (((v[1] + ((v[1] << 4) ^ (v[1] >> 5))) + k4) * 9,
                       rot_box3[idx2]);
              v[2] -=
                rol32 (((v[0] + ((v[0] << 4) ^ (v[0] >> 5))) + k3) * 9,
                       rot_box2[idx1]);
            }
          else
            {
              v[3] ^=
                rol32 (((v[1] + ((v[1] << 4) ^ (v[1] >> 5))) + k4) * 9,
                       rot_box3[idx2]);
              v[2] ^=
                rol32 (((v[0] + ((v[0] << 4) ^ (v[0] >> 5))) + k3) * 9,
                       rot_box2[idx1]);
            }



          if (idx2 == idx2null)
            {
              v[1] -=
                (v[2] + ror32 ((v[0] + ((v[0] << 6) ^ (v[0] >> 9))), rot2) +
                 k2);
            }
          else
            {
              v[1] -=
                (v[2] +
                 ror32 ((v[0] + ((v[0] << 6) ^ (v[0] >> 9))) + k2, rot2));
            }



          if (idx1 == idx1null)
            {
              v[0] -=
                (v[3] + ror32 ((v[1] + ((v[1] << 6) ^ (v[1] >> 9))), rot1) +
                 k1);
            }
          else
            {
              v[0] -=
                (v[3] +
                 ror32 ((v[1] + ((v[1] << 6) ^ (v[1] >> 9))) + k1, rot1));
            }


          v[0] =
            (BackSBox[idx1][idx2][(u8) (v[0] >> 0)] << 0) +
            (BackSBox[idx1][idx2][(u8) (v[0] >> 8)] << 8) +
            (BackSBox[idx1][idx2][(u8) (v[0] >> 16)] << 16) +
            (BackSBox[idx1][idx2][(u8) (v[0] >> 24)] << 24);

          v[1] =
            (BackSBox[idx1 + 1][idx2 + 1][(u8) (v[1] >> 0)] << 0) +
            (BackSBox[idx1 + 1][idx2 + 1][(u8) (v[1] >> 8)] << 8) +
            (BackSBox[idx1 + 1][idx2 + 1][(u8) (v[1] >> 16)] << 16) +
            (BackSBox[idx1 + 1][idx2 + 1][(u8) (v[1] >> 24)] << 24);

          v[2] =
            (BackSBox[idx1][idx2 + 1][(u8) (v[2] >> 0)] << 0) +
            (BackSBox[idx1][idx2 + 1][(u8) (v[2] >> 8)] << 8) +
            (BackSBox[idx1][idx2 + 1][(u8) (v[2] >> 16)] << 16) +
            (BackSBox[idx1][idx2 + 1][(u8) (v[2] >> 24)] << 24);

          v[3] =
            (BackSBox[idx1 + 1][idx2][(u8) (v[3] >> 0)] << 0) +
            (BackSBox[idx1 + 1][idx2][(u8) (v[3] >> 8)] << 8) +
            (BackSBox[idx1 + 1][idx2][(u8) (v[3] >> 16)] << 16) +
            (BackSBox[idx1 + 1][idx2][(u8) (v[3] >> 24)] << 24);


          if (idx1 != 0)
            {
              idx1--;
            }
          else
            {
              idx1 = (subkeysize - 2);
            }

          if (idx2 != 0)
            {
              idx2--;
            }
          else
            {
              idx2 = (subkeysize - 2);
            }


          s = v[0] ^ v[1] ^ v[2] ^ v[3];

          for (j = 3; j != -1; j--)
            {
              m = (s >> (8 * j)) & 3;
              t = v[j];
              v[j] = v[m];
              v[m] = t;
            }



        }
    }
}
